#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#pylint: disable-msg=E1101,C0103,R0902,R0903
#Author: Valentin Kuznetsov

"""
ReleaseManager is responsible for setting up CMSSW area and run cmsRun job 
to extract run/events from given LFN/PFN list
"""

import os
import time
import stat
import popen2
import tempfile
import threading
import ConfigParser
from   FileMover import ConfiguredObject
from   FileManager import ThreadPool

class CMSRunManager(object):

    def __init__(self, area, release, config):
        self.area = area         # CMSSW release area to be used, setup by scram
        self.release = release   # CMSSW release to be used
        self.config  = config    # cmsRun configuration
        self.process = None      # cmsRun process Popen4 object
        self.tmpfilename = None  # cmsRun configuration file name
        self.release_to_use = '' # the CMSSW released used to produce output file
        self.outfilename    = '' # output file name
        self.inputparams = None  # input parameters: dataset, eventset

        # find list of available release on our system
        cmd = "scram list -c CMSSW | awk '{print $2}'"
        res = popen2.Popen4(cmd)
        res.wait()
        releases = res.fromchild.read()
        self.releases = releases.split()
        
        # find best matched release wrt one we need to use
        for rel in self.releases:
            if  rel < self.release:
                self.release_to_use = rel

    def setup_cmssw(self):
        cdir = os.getcwd()
        os.chdir(self.area)
        cmd = 'scramv1 project CMSSW %s' % self.release_to_use
        os.system(cmd)
        os.chdir(cdir)

    def start(self):
        cdir = os.getcwd()
        self.setup_cmssw()
        os.chdir( os.path.join( self.area, self.release_to_use ))
        if  self.release_to_use < 'CMSSW_1_6':
            return # no support
        elif self.release_to_use < 'CMSSW_2_1':
            tmpfilename = tempfile.mktemp() + '.cfg'
        else:
            tmpfilename = tempfile.mktemp() + '.py'
        self.tmpfilename = tmpfilename
        tfile = open(tmpfilename, 'w')
        tfile.write(self.config)
        tfile.close()
        cmd = 'eval `scramv1 run -sh` && cmsRun %s' % tmpfilename
        print cmd
#        print self.config
#        os.system(cmd)
        process = popen2.Popen4(cmd)
        self.process = process
#        print "cmsRun process:", self.process, type(self.process), self.process.pid, self.process.poll()
#        os.remove(tmpfilename)
        os.chdir(cdir)
        
    def process_output(self):
        if  self.process:
            self.process.wait()
            res = self.process.fromchild.read()
            return res
        return
        
    def cancel(self):
        # implement kill -9
        if  self.process:
            pid = self.process.pid
            os.system('kill -9 %s' % pid)
            self.process = None
            os.remove(self.tmpfilename)
        return

    def status(self):
        status = -1
        if  self.process:
            status = self.process.poll()
            if  not status:
                self.process = None
                os.remove(self.tmpfilename)
        return status

    def filecontent(self):
        cwd  = os.getcwd()
        area = os.path.join(self.area, self.release_to_use)
        os.chdir(area)
        cmd  = 'eval `scramv1 run -sh` && edmFileUtil -e -f file:%s' % self.outfilename
        res  = popen2.Popen4(cmd)
        res.wait()
        content = res.fromchild.read()
        os.chdir(cwd)
        return content
    
class CMSRunThreadWrapper(ConfiguredObject):
    """
       this class handles setup of CMSSW area and running cmsRun job
    """
    def __init__(self, area):

        self.cp       = None
        self.section  = "release_manager"
        super(CMSRunThreadWrapper, self).__init__(area)
        self._request = {}
        self.max_cmsrun = 3
        self.request_lock = threading.Lock()

    def configure(self, cp):
        self.cp = cp
        self.max_cmsrun = int(self.getOption("max_cmsrun", "5"))
        self.pool = ThreadPool(threads=self.max_cmsrun)

    def request(self, area, release, config):
        """
           run cmsRun job over provided config file, returns output root file
        """
        if  len(self._request.keys()) == self.max_cmsrun:
            return None # to prevent from having too many cmsrun jobs on a node
        requestid = str(time.time())
        self.request_lock.acquire()
        try:
            try:
                cmsrun = CMSRunManager(area, release, config)
                self._request[requestid] = cmsrun
                self.pool.queue(cmsrun)
            except Exception, e:
                # unable to start new thread
                self.log.exception(e)
                requestid = None
                pass
        finally:
            self.request_lock.release()
        return requestid

    def status(self, id):
        """
           return status of cmsRun job
        """
        status = -1
        if  not id:
            return 'busy'
        self.request_lock.acquire()
        try:
            try:
                cmsrun = self._request[id]
                status = cmsrun.status()
                return status
            except (SystemExit, KeyboardInterrupt):
                raise
            except Exception, e:
                self.log.exception(e)
                return status
        finally:
            self.request_lock.release()
        return status
        
    def updateparams(self, id, outfilename, params):
        """update object with given input parameters for given request id"""
        self._request[id].outfilename = outfilename
        self._request[id].inputparams = params
        # look-up for old files and clean them up
        # delete files older then 3 hours
        odir = os.path.split(outfilename)[0]
        for file in os.listdir(odir):
            ofile = os.path.join(odir, file)
            filestat = os.stat(ofile)
            if  filestat[stat.ST_MTIME] > (60*60*3):
                os.remove(ofile)
        
    def inputparams(self, id):
        """return input parameters for given request id"""
        return self._request[id].inputparams
        
    def getoutfile(self, id):
        """return output file name for given request id"""
        return self._request[id].outfilename
        
    def filecontent(self, id):
        """return file content of request id"""
        return self._request[id].filecontent()
        
    def release_to_use(self, id):
        """return file content of request id"""
        return self._request[id].release_to_use
        
    def config(self, id):
        """return file content of request id"""
        return self._request[id].config

    def process_output(self, id):
        """return output of cmsRun for request id"""
        return self._request[id].process_output()
        
    def clean(self, id):
        """clean internal cache about request"""
        del self._request[id]
        
    def cancel(self, id):
        self.request_lock.acquire()
        try:
            cmsrun = self._request[id]
            self.log.info("Sending a cancel request to cmsrun request=%s, filename=%s" \
                                % (id, cmsrun.outfilename))
            cmsrun.cancel()
            self.clean(id)
        finally:
            self.request_lock.release()

    def graceful_exit(self):
        """implement exit from thread pool"""
        self.pool.drain()
        self.pool.join()
#
# main
#
def test(dbsmgr, mgr, dir, dataset, run, evt, lumi):
    files = dbsmgr.getFiles(run, dataset, evt, lumi)
    if  files:
        release = dbsmgr.getRelease(files[0][0])
        opath = dir + '/root_files/valya'
        config, outfilename = dbsmgr.edmconfig(dir, dataset, run, evt, lumi)
        requestid = mgr.request(dir, release, config)
        if  not requestid:
            print "#### UNABLE to start thread for", dataset, run, evt, lumi
        return requestid
    return None, None

def check_status(mgr, requestid):
    counter = 0
    outfilename = mgr.getoutfile(requestid)
    import time, stat
    while 1:
        status = mgr.status(requestid)
        if  status == 'busy':
            break
        if  os.path.isfile(outfilename):
            print '%s %s' % (os.stat(outfilename)[stat.ST_SIZE], outfilename)
        else:
            print 'cmsRun job is launched, please wait, job id: %s, status %s' % (outfilename, status)
        if  not status:
            print "Finish!!!", outfilename
            break
        counter += 1
        time.sleep(3)
        if counter > 25:
            break

def main():
    dir = '/data/vk/filemover/cmssw'
    mgr = CMSRunThreadWrapper(dir)
    cp = ConfigParser.ConfigParser()
    cp.add_section("release_manager")
    cp.set("release_manager", "max_cmsrun", "2")
    mgr.configure(cp)
    import thread
    import DBSInteraction
    dbsmgr = DBSInteraction.DBS('cms_dbs_prod_global')
    # test CMSSW_1_6_X release series
    dataset = '/testbeam_HCalEcalCombined/h2tb2007_default_v1/DIGI-RECO'
    run = 16278
    evt = 44000
    lumi = 10
    requestid = test(dbsmgr, mgr, dir, dataset, run, evt, lumi)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    # test CMSSW_2_1_X release series
    dataset = '/Cosmics/Commissioning08-CRUZET4_v1/RECO'
    run = 58620
    evt = 151578
    lumi = 6
    requestid = test(dbsmgr, mgr, dir, dataset, run, evt, lumi)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    # use the same dataset but real event, this time we should fail since 
    # we only allow 2 cmsRun jobs
    evt = 10
    requestid = test(dbsmgr, mgr, dir, dataset, run, evt, lumi)
    print "####", requestid, mgr.getoutfile(requestid)
    if  requestid:
        thread.start_new_thread(check_status, (mgr, requestid))
    time.sleep(90)

if __name__ == "__main__":
    main()
