#!/usr/bin/env python
#-*- coding: ISO-8859-1 -*-
#
# Author: Brain Bockelman

import os
import re
import time
import signal
import urllib
import urllib2
import logging
import threading
import xml.dom.minidom

from Utils import LfnInfoCache, getPercentageDone
from Status import StatusMsg, StatusCode
from DBSInteraction import DBS

logging.basicConfig(level=logging.INFO)

def json(data_str):
    return eval(data_str, {'null': None}, {})

def phedex_datasvc(query, db='prod',
        endpoint='http://cmsweb.cern.ch/phedex/datasvc', **kw):
    get_params = urllib.urlencode(kw)
    url = os.path.join(endpoint, 'json', db, query)
    url += "?" + get_params
    try:
        results = urllib2.urlopen(url).read()
    except:
        raise Exception("Failed to contact the PhEDEx datasvc")
    try:
        return json(results)
    except:
        raise Exception("PhEDEx datasvc returned an error.")

class ConfiguredObject(object):

    def __init__(self, *args, **kw):
        self.log = logging.getLogger(self.__class__.__name__)

    def getOption(self, option, default=None):
        try:
            return self.cp.get(self.section, option)
        except:
            return default

class FileMover(ConfiguredObject):

    def __init__(self, cp):
        super(FileMover, self).__init__(cp)
        self.cp = cp
        self.section = "file_mover"
        self.lookup_object = FileLookup(cp)
        self.source = None
        self.transfer_wrapper = None
        self.lfn = None
        self.is_cached = False

    def request(self, lfn, dest_dir):
        self.lfn = lfn
        self.dest_dir = dest_dir
        dest_dir = os.path.join(dest_dir, os.path.split(lfn)[0][1:])
        self._create_dest_dir(dest_dir)
        self.source = self.lookup_object.getPFN(self.lfn)

    def getLFN(self):
        return self.lfn

    def _create_dest_dir(self, dir):
        try:
            os.makedirs(dir)
            self.log.info("Created directory %s" % dir)
        except Exception, e: 
            if str(e).find('Errno 17') < 0:
                raise e

    def check_cache(self):
        local_pfn = os.path.join(self.dest_dir, self.lfn[1:])
        if os.path.exists(local_pfn):
            return True
        return False

    def start(self):
        if not self.lfn:
            raise Exception("You must first request a file!")
        if self.transfer_wrapper:
            raise Exception("Transfer has already been launched!")
        else:
            if self.check_cache():
                self.is_cached = True
                return
            local_pfn = os.path.join(self.dest_dir, self.lfn[1:])
            dest = 'file:///' + local_pfn
            self.transfer_wrapper = TransferWrapper(self.cp, self.source, dest)
            self.transfer_wrapper.launch()

    def cancel(self):
        if self.transfer_wrapper:
            self.log.info("Passing a cancel request from the file mover to the"\
                " transfer wrapper.")
            self.transfer_wrapper.cancel()
        else:
            self.log.warning("Trying to cancel a transfer which hasn't" \
                " started.")

    def status(self):
        if self.is_cached:
            return (StatusCode.DONE, StatusMsg.OBJECT_IN_CACHE)
        if self.transfer_wrapper:
            return self.transfer_wrapper.status()
        return (StatusCode.TRANSFER_WRAPPER_NOT_LAUNCHED,
            StatusMsg.TRANSFER_WRAPPER_NOT_LAUNCHED)
        
class FileLookup(ConfiguredObject):

    def __init__(self, cp):
        super(FileLookup, self).__init__(cp)
        self.cp = cp
        self.section = "file_lookup"
        self.map_base = self.getOption("tfc_service", \
            "http://t2.unl.edu/phedex/tfc/map")
        self.replica_base = self.getOption("replica_service", \
            "http://t2.unl.edu/phedex/xml/replica_info")
        self.priorities = self._parse_priority_rules()
        self._dbs = DBS()

    def replicas(self, lfn):
        self.log.info("Looking for the block of %s." % lfn)
        block = self._dbs.blockLookup(lfn)
        query = {'block':block}
        self.log.info("Looking for replicas of %s" % block)
        results = phedex_datasvc('fileReplicas', **query)
        blocks = [i for i in results['phedex']['block'] if i.get('name', None) \
            == block]
        if not blocks:
            raise Exception("Requested LFN does not exist in any block known " \
                "to PhEDEx.")
        block = blocks[0]
        files = [i for i in block.get('file', []) if i.get('name', None) == \
            lfn]
        if not files:
            raise Exception("Internal error: PhEDEx does not think LFN is in " \
                "the same block as DBS does.")
        file = files[0]
        replicas = [i['node'] for i in file.get('replica', []) if 'node' in i]
        self.log.info("There are the following replicas of %s: %s." % \
            (lfn, ', '.join(replicas)))
        return replicas

    def _parse_priority_rules(self):
        priority_dict = {} 
        priority_dict[1e6-1] = re.compile('T1_US')
        priority_dict[1e6] = re.compile('T1')
        priority_dict[1e6+1] = re.compile('T2')
        priority_dict[1e6+2] = re.compile('.*')
        name_regexp = re.compile('priority_([0-9]+)')
        try:
            rules = self.cp.items(self.section)
        except:
            rules = {}
        for name, value in rules:
            value = value.strip()
            value = re.compile(value)
            m = name_regexp.match(name)
            if not m:
               continue
            priority = long(m.groups()[0])
            priority_dict[priority] = value
        return priority_dict

    def pickSite(self, replicas):
        priorities = self.priorities.keys()
        priorities.sort()
        source = None
        for priority in priorities:
            for site in replicas:
                m = self.priorities[priority].search(site)
                if m:
                    source = site
                    break
            if source != None:
                break
        if source == None:
            raise ValueError("Could not match site to any priority.  " \
                "Possible sources: %s" % str(replicas))
        return source

    def mapLFN(self, site, lfn, protocol='srmv2'):
        data = {'lfn': lfn, 'node': site}
        if protocol:
            data['protocol'] = protocol
        self.log.info("Mapping LFN %s for site %s using PhEDEx datasvc." % \
            (lfn, site))
        data = phedex_datasvc('lfn2pfn', **data)
        try:
            pfn = data['phedex']['mapping'][0]['pfn']
        except:
            raise Exception("PhEDEx data service did not return a PFN!")
        return pfn

    def getPFN(self, lfn):
#        replicas = self.replicas(lfn)
#        if len(replicas) == 0:
#            raise Exception("The LFN %s has no known replicas in PhEDEx.")
#        site = self.pickSite(replicas)
#        pfn = self.mapLFN(site, lfn)
#        return pfn

        site = ''
        try:
            replicas = self.replicas(lfn)
            if len(replicas) == 0:
                raise \
                Exception("The LFN %s has no known replicas in PhEDEx."%lfn)
            site = self.pickSite(replicas)
        except:
            bsList = self._dbs.blockSiteLookup(lfn)
            seList = [s for b,s in bsList]
            site = self.getSiteFromSDB(seList)
            if  not site:
                msg = "Fail to look-up CMS name for\n"
                msg+= "LFN=%s\n"%lfn
                msg+= "SE's list %s\n"%seList
                msg+= "Please check SiteDB that those SE's are registered."
                raise Exception(msg)
        pfn = self.mapLFN(site, lfn)
        return pfn

    def getSiteFromSDB(self,seList):
        # try another route, get SE from DBS, look-up CMS name from SiteDB
        site = ''
        for se in seList:
            url = "https://cmsweb.cern.ch/sitedb/json/index/SEtoCMSName"
            values = {'name':se}
            data = urllib.urlencode(values)
            req  = urllib2.Request(url,data)
            response = urllib2.urlopen(req)
            the_page = response.read().replace('null','None')
            json = eval(compile(the_page, '<string>', 'eval'))
            mylist = [i['name'] for i in json.values() \
                                    if i['name'].find('T0_')==-1]
            mylist.sort()
            site = mylist[-1]
            break
        # Do some magic for T1's. SiteDB returns names as T1_US_FNAL, while
        # phedex will need name_Buffer, according to Simon there is a 
        # savannah ticket for that.
        if site.count('T1',0,2) == 1:
            if site.count('Buffer') == 0 and site.count('Buffer') ==0:
                site = "%s_Buffer" % site
        return site

class TransferWrapper(ConfiguredObject):

    def __init__(self, cp, source, dest):
        self.cp = cp
        self.section = "transfer_wrapper"
        super(TransferWrapper, self).__init__(cp)
        self.pid = None
        self.source = source
        self.dest = dest
        self._killflag = False
        self.log.info("Transfer from %s to %s." % (source, dest))
        self.final_status = None
        self.lfnInfoCache = LfnInfoCache()

    def launch(self):
        self._launch_process()
        status = self.status()[0]
        process_running_codes = [2]
        while status in process_running_codes and not self._killflag:
            try:
                time.sleep(3)
                status = self.status()
                self.log.info("Current transfer status: %s." % str(status[1]))
                status = status[0]
            except:
                self.cancel()
        if self._killflag:
            self.log.info("Transfer wrapper exiting due to kill flag.")

    def _launch_process(self, blocking=False):
        if blocking:
            options = os.P_WAIT
        else:
            options = os.P_NOWAIT
        srmcp_command = self.getOption("transfer_command",
            "srmcp -debug=true -use_urlcopy_script=true " \
            "-srm_protocol_version=2 -retry_num=1")
        srmcp_args = srmcp_command.split()
        srmcp_args = [srmcp_args[0]] + srmcp_args
        srmcp_args += [self.source, self.dest]
        self.log.info("Launching command %s." % ' '.join(srmcp_args))
        # We wrap this with a simple python script which sets the process
        # group for the child process, then launches srmcp.  This means we
        # can later send a signal to the entire process group, killing srmcp's
        # children processes too, instead of just killing the wrapper.
        srmcp_args = ["/usr/bin/python", "srmcp_wrapper", "-c",
            "import os, sys; os.setpgrp(); os.execvp(sys.argv[1]," \
            " sys.argv[2:])"] + srmcp_args
        results = os.spawnlp(options, *srmcp_args)
        if blocking:
            self.status = results
        else:
            self.pid = results

    def process_status(self):
        if not self.pid:
            raise Exception("Process was not started!")
        status = os.waitpid(self.pid, os.WNOHANG)
        if status == (0, 0): # No status available
            return None
        if os.WIFSIGNALED(status[1]):
            return -os.WTERMSIG(status[1])
        elif os.WIFEXITED(status[1]):
            return os.WEXITSTATUS(status[1])
        raise Exception("Unable to determine job status!")

    def file_progress_status(self):
        if self.dest.startswith('file:///'):
            dest = self.dest[7:]
        else:
            dest = self.dest
        try:
            stat = os.stat(dest)
        except OSError, oe:
            if oe.errno == 2:
#                return "SRM portion of transfer still going"
                return StatusMsg.WAITING_FOR_SRM
            else:
                raise
        size = stat[6]
        if size == 0:
            return StatusMsg.GRIDFTP_NO_MOVEMENT
        else:
            perc = ""
            try: 
               baseDir = self.cp.get('file_manager', 'base_directory')
               myLfn = dest.replace("file://", "").replace(baseDir, "")
               perc = "%s%%," % getPercentageDone(size,
                   self.lfnInfoCache.getSize(myLfn))
            except:
#               import traceback
#               traceback.print_exc()
               pass
            return StatusMsg.IN_PROGRESS % (perc, round(size/1024.0**2))

    def status(self):
        if self.final_status:
            return self.final_status
        if not self.pid:
            return  (StatusCode.TRANSFER_PROCESS_NOT_STARTED,
                StatusMsg.TRANSFER_PROCESS_NOT_STARTED)
        process_status = self.process_status()
        if process_status == None:
            return (2, self.file_progress_status())
        elif process_status == 0:
            self.final_status = (StatusCode.DONE, StatusMsg.FILE_DONE)
            return self.final_status
        else:
            self.final_status = (StatusCode.TRANSFER_FAILED,
                StatusMsg.TRANSFER_FAILED_STATUS % process_status)
            return self.final_status
        return (StatusCode.TRANSFER_STATUS_UNKNOWN,
                StatusMsg.TRANSFER_STATUS_UNKNOWN)

    def cancel(self):
        self.log.info("Starting the cancel of transfer_wrapper %s" % self)
        if self.dest.startswith('file:///'):
            dest = self.dest[7:]
        else:
            dest = self.dest
        if os.path.exists(dest):
            self.log.info("Unlinking partially complete dest file %s." % dest)
            os.unlink(dest)
        else:
            self.log.info("Destination path %s doesn't exist; not deleting." % \
                dest)
        self._killflag = True
        if self.pid:
            self.log.info("Killing transfer process at PID %s." % str(self.pid))
            os.killpg(self.pid, signal.SIGTERM)
            self.log.info("Process return status: %s." % \
                str(os.waitpid(self.pid, os.P_WAIT)))
            self.pid = None
        else:
            self.log.warning("I don't know what PID to kill!  Doing nothing.")
        self.log.info("Setting the kill flag, which should cause the " \
            "transfer_wrapper to exit soon.")

